package ru.dbotthepony.mc.otm.block.entity;

import net.minecraft.core.BlockPos;
import net.minecraft.core.Direction;
import net.minecraft.nbt.CompoundTag;
import net.minecraft.world.item.ItemStack;
import net.minecraft.world.level.block.entity.BlockEntityType;
import net.minecraft.world.level.block.state.BlockState;
import net.minecraftforge.common.capabilities.Capability;
import net.minecraftforge.common.util.LazyOptional;
import net.minecraftforge.energy.CapabilityEnergy;
import net.minecraftforge.energy.IEnergyStorage;
import ru.dbotthepony.mc.otm.capability.IMatteryEnergyStorage;
import ru.dbotthepony.mc.otm.capability.MatteryCapability;
import ru.dbotthepony.mc.otm.capability.MatteryMachineEnergyStorage;
import ru.dbotthepony.mc.otm.container.SimpleSerializableContainer;

import javax.annotation.Nonnull;
import javax.annotation.Nullable;
import java.math.BigDecimal;
import java.util.Optional;

abstract public class BlockEntityMatteryPowered extends BlockEntityMattery {
	protected MatteryMachineEnergyStorage energy = null;
	protected final LazyOptional<MatteryMachineEnergyStorage> energy_resolver = LazyOptional.of(() -> energy);;
	private boolean valid = true;

	public SimpleSerializableContainer battery_container = new SimpleSerializableContainer(this::setChanged, 1);

	protected double workSpeedMultiplier(BigDecimal energy_required, boolean simulate) {
		BigDecimal extract_much = energy.extractEnergyInner(energy_required, simulate);
		return extract_much.divide(energy_required, MatteryCapability.ROUND_RULES).doubleValue();
	}

	protected double workSpeedMultiplier(BigDecimal energy_required) {
		return workSpeedMultiplier(energy_required, false);
	}

	protected void tickBatteryCharge() {
		if (energy == null)
			return;

		if (energy.getMissingPower().compareTo(BigDecimal.ZERO) <= 0)
			return;

		BigDecimal demand = energy.getMissingPower();
		demand = demand.min(energy.receiveEnergyOuter(demand, true));

		for (int i = 0; i < battery_container.getContainerSize() && demand.compareTo(BigDecimal.ZERO) > 0; i++) {
			ItemStack stack = battery_container.getItem(i);

			if (!stack.isEmpty()) {
				Optional<IMatteryEnergyStorage> mattery_storage = stack.getCapability(MatteryCapability.ENERGY).resolve();

				if (mattery_storage.isPresent()) {
					IMatteryEnergyStorage storage = mattery_storage.get();

					BigDecimal drain = storage.extractEnergyOuter(demand, true);

					if (drain.compareTo(BigDecimal.ZERO) != 0) {
						BigDecimal receive = energy.receiveEnergyOuter(drain, true);

						storage.extractEnergyOuter(receive, false);
						energy.receiveEnergyOuter(receive, false);

						demand = demand.subtract(receive, MatteryCapability.ROUND_RULES);
					}
				} else if (demand.compareTo(BigDecimal.ONE) >= 0) {
					Optional<IEnergyStorage> storage = stack.getCapability(CapabilityEnergy.ENERGY).resolve();

					if (storage.isPresent()) {
						IEnergyStorage energy_fe = storage.get();

						BigDecimal drain = MatteryCapability.drainFE(energy_fe, demand, true);

						if (drain.compareTo(BigDecimal.ZERO) != 0) {
							BigDecimal receive = energy.receiveEnergyOuter(drain, true);

							MatteryCapability.drainFE(energy_fe, receive, false);
							energy.receiveEnergyOuter(receive, false);

							demand = demand.subtract(receive, MatteryCapability.ROUND_RULES);
						}
					}
				}
			}
		}
	}

	public BlockEntityMatteryPowered(BlockEntityType<?> p_155228_, BlockPos p_155229_, BlockState p_155230_) {
		super(p_155228_, p_155229_, p_155230_);
	}

	@Override
	public void invalidateCaps() {
		super.invalidateCaps();
		valid = false;
	}

	@Override
	public void reviveCaps() {
		super.reviveCaps();
		valid = true;
	}

	@Nonnull
	public <T> LazyOptional<T> getCapability(@Nonnull Capability<T> cap, @Nullable Direction side) {
		if (valid && (cap == MatteryCapability.ENERGY || cap == CapabilityEnergy.ENERGY) && energy != null) {
			return energy_resolver.cast();
		}

		return super.getCapability(cap, side);
	}

	@Override
	public CompoundTag save(CompoundTag nbt) {
		nbt.put("energy_cap", energy.serializeNBT());
		nbt.put("battery_container", battery_container.serializeNBT());

		return super.save(nbt);
	}

	public void load(CompoundTag nbt) {
		if (nbt.contains("energy_cap") && nbt.get("energy_cap") instanceof CompoundTag tag)
			energy.deserializeNBT(tag);

		if (nbt.contains("battery_container") && nbt.get("battery_container") instanceof CompoundTag tag)
			battery_container = SimpleSerializableContainer.of(this::setChanged, tag, 1);

		super.load(nbt);
	}
}
